@@ -34,6 +34,17 @@ module Hpricot
       output("", :preserve => true)
     end
 
+    def index(name)
+      i = 0
+      return i if name == "*"
+      children.each do |x|
+        return i if (x.respond_to?(:name) and name == x.name) or
+          (x.text? and name == "text()")
+        i += 1
+      end
+      -1
+    end
+
     # Puts together an array of neighboring nodes based on their proximity
     # to this node.  So, for example, to get the next node, you could use
     # <tt>nodes_at(1).  Or, to get the previous node, use <tt>nodes_at(1)</tt>.
@@ -46,11 +57,19 @@ module Hpricot
     def nodes_at(*pos)
       sib = parent.children
       i, si = 0, sib.index(self)
+      pos.map! do |r|
+        if r.is_a?(Range) and r.begin.is_a?(String)
+          r = Range.new(parent.index(r.begin)-si, parent.index(r.end)-si, r.exclude_end?)
+        end
+        r
+      end
+      p pos
       Elements[*
         sib.select do |x|
-          sel = case i - si when *pos
-                  true
-                end
+          sel =
+            case i - si when *pos
+              true
+            end
           i += 1
           sel
         end
@@ -59,35 +78,51 @@ module Hpricot
 
     # Returns the node neighboring this node to the south: just below it.
     # This method includes text nodes and comments and such.
-    def next_node
+    def next
       sib = parent.children
       sib[sib.index(self) + 1] if parent
     end
+    alias_method :next_node, :next
 
     # Returns to node neighboring this node to the north: just above it.
     # This method includes text nodes and comments and such.
-    def previous_node
+    def previous
       sib = parent.children
       x = sib.index(self) - 1
       sib[x] if sib and x >= 0
     end
+    alias_method :previous_node, :previous
+
+    # Find all preceding nodes.
+    def preceding
+      sibs = parent.children
+      si = sibs.index(self) 
+      return Elements[*sibs[0...si]] 
+    end 
+ 
+    # Find all nodes which follow the current one.
+    def following
+      sibs = parent.children 
+      si = sibs.index(self) + 1 
+      return Elements[*sibs[si...sibs.length]] 
+    end 
 
     # Adds elements immediately after this element, contained in the +html+ string.
-    def after(html)
-      parent.insert_after(Hpricot.make(html), self)
+    def after(html = nil, &blk)
+      parent.insert_after(Hpricot.make(html, &blk), self)
     end
 
     # Adds elements immediately before this element, contained in the +html+ string.
-    def before(html)
-      parent.insert_after(Hpricot.make(html), self)
+    def before(html = nil, &blk)
+      parent.insert_before(Hpricot.make(html, &blk), self)
     end
 
 
     # Replace this element and its contents with the nodes contained
     # in the +html+ string.
-    def swap(html)
+    def swap(html = nil, &blk)
       parent.altered!
-      parent.replace_child(self, Hpricot.make(html))
+      parent.replace_child(self, Hpricot.make(html, &blk))
     end
 
     def get_subnode(*indexes)
@@ -116,26 +151,29 @@ module Hpricot
     alias_method :innerText, :inner_text
 
     # Builds an HTML string from the contents of this node.
-    def inner_html
-      if respond_to? :children
-        children.map { |x| x.output("") }.join
+    def html(inner = nil, &blk)
+      if inner or blk
+        altered!
+        case inner
+        when Array
+          self.children = inner
+        else
+          self.children = Hpricot.make(inner, &blk)
+        end
+        reparent self.children
+      else
+        if respond_to? :children
+          children.map { |x| x.output("") }.join
+        end
       end
     end
+    alias_method :inner_html, :html
     alias_method :innerHTML, :inner_html
 
     # Inserts new contents into the current node, based on
     # the HTML contained in string +inner+.
     def inner_html=(inner)
-      altered!
-      case inner
-      when String, IO
-        self.children = Hpricot.parse(inner).children
-      when Array
-        self.children = inner
-      when nil
-        self.children = []
-      end
-      reparent self.children
+      html(inner || [])
     end
     alias_method :innerHTML=, :inner_html=
 
@@ -197,6 +235,9 @@ module Hpricot
     # containing the matching nodes.  If +blk+ is given, it
     # is used to iterate through the matching set.
     def search(expr, &blk)
+      if Range === expr
+        return Elements.expand(at(expr.begin), at(expr.end), expr.exclude_end?)
+      end
       last = nil
       nodes = [self]
       done = []
@@ -210,7 +251,7 @@ module Hpricot
           when %r!^/?\.\.!
               last = expr = $'
               nodes.map! { |node| node.parent }
-          when %r!^[>/]!
+          when %r!^[>/]\s*!
               last = expr = $'
               nodes = Elements[*nodes.map { |node| node.children if node.respond_to? :children }.flatten.compact]
           when %r!^\+!
@@ -364,6 +405,22 @@ module Hpricot
       sib[x] if sib and x >= 0
     end
 
+    # Find all preceding sibling elements.   Like the other "sibling" methods, this weeds
+    # out text and comment nodes.
+    def preceding_siblings() 
+      sibs = parent.containers 
+      si = sibs.index(self) 
+      return Elements[*sibs[0...si]] 
+    end 
+ 
+    # Find sibling elements which follow the current one.   Like the other "sibling" methods, this weeds
+    # out text and comment nodes.
+    def following_siblings() 
+      sibs = parent.containers 
+      si = sibs.index(self) + 1 
+      return Elements[*sibs[si...sibs.length]] 
+    end 
+
     # Puts together an array of neighboring sibling elements based on their proximity
     # to this element.
     #
@@ -412,7 +469,7 @@ module Hpricot
     def insert_after(nodes, ele)
       case nodes
       when Array
-        nodes.each { |n| insert_after(n, ele) }
+        nodes.reverse_each { |n| insert_after(n, ele) }
       else
         reparent nodes
         idx = children.index(ele)
@@ -508,7 +565,7 @@ module Hpricot
     # +each_hyperlink_uri+ traverses hyperlinks such as HTML href attribute
     # of A element.
     #
-    # It yields Hpricot::Text (or Hpricot::Loc) and URI for each hyperlink.
+    # It yields Hpricot::Text and URI for each hyperlink.
     #
     # The URI objects are created with a base URI which is given by
     # HTML BASE element or the argument ((|base_uri|)).
@@ -533,7 +590,7 @@ module Hpricot
     # +each_hyperlink+ traverses hyperlinks such as HTML href attribute
     # of A element.
     #
-    # It yields Hpricot::Text or Hpricot::Loc.
+    # It yields Hpricot::Text.
     #
     # Note that +each_hyperlink+ yields HTML href attribute of BASE element.
     def each_hyperlink # :yields: text
@@ -738,22 +795,25 @@ module Hpricot
 
   module Elem::Trav
     def has_attribute?(name)
-      self.attributes && self.attributes.has_key?(name.to_s)
+      self.raw_attributes && self.raw_attributes.has_key?(name.to_s)
     end
     def get_attribute(name)
-      self.attributes && self.attributes[name.to_s]
+      a = self.raw_attributes && self.raw_attributes[name.to_s]
+      a = Hpricot.uxs(a) if a
+      a
     end
     alias_method :[], :get_attribute
     def set_attribute(name, val)
       altered!
-      self.attributes ||= {}
-      self.attributes[name.to_s] = val
+      self.raw_attributes ||= {}
+      self.raw_attributes[name.to_s] = Hpricot.xs(val)
     end
     alias_method :[]=, :set_attribute
     def remove_attribute(name)
+      name = name.to_s
       if has_attribute? name
         altered!
-        self.attributes.delete(name)
+        self.raw_attributes.delete(name)
       end
     end
   end
